// StopwatchView.cpp : implementation of the CStopwatchView class
//

#include "stdafx.h"
#include "Stopwatch.h"

#include "mmsystem.h"

#include "StopwatchDoc.h"
#include "StopwatchView.h"
#include ".\stopwatchview.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#endif

#define HISTORY_LENGTH 10000	// maximum number of events to be recorded

WAVEFORMATEX WaveInFormat;			// wave input format
HWAVEIN HandleWaveIn;
WAVEHDR WaveHeaderIn1;				// wave input header
WAVEHDR WaveHeaderIn2;				// wave input header
short (*InputBuffer1)[2];			// wave input buffer
short (*InputBuffer2)[2];			// wave input buffer
int BufferLength;					// length of the buffer for raw sound card data
int SampleRate;						// sample rate can be 11025Hz, 22050Hz or 44100Hz
int NumberOfEvents;					// number of events already captured
int HistoryData[HISTORY_LENGTH];	// recorded event time instants


int MeasurementStarted=0;			// indicates if a measurement has been started
int PreviousSample;					// holds the previous measured sample, used to calculate threshold crossing

int Threshold=16384;				// threshold to detect events (16384 is half of the signal range)
int ThresholdTolerance=1000;		// this is the hystheresis
int Counter;						// counter is started and stopped at subsequent events
int BlockingCounter;				// this counter runs to detect the duration of the light beam blocking - not yet used
int TimerIsRunning;					// indicates that the timer is running (counter is counting)
int GateIsBlocked;					// indicates if the light beam is blocked


// CStopwatchView

IMPLEMENT_DYNCREATE(CStopwatchView, CView)

BEGIN_MESSAGE_MAP(CStopwatchView, CView)
	// Standard printing commands
	ON_COMMAND(ID_FILE_PRINT, &CView::OnFilePrint)
	ON_COMMAND(ID_FILE_PRINT_DIRECT, &CView::OnFilePrint)
	ON_COMMAND(ID_FILE_PRINT_PREVIEW, &CView::OnFilePrintPreview)
	ON_COMMAND(ID_MEASUREMENT_START, &CStopwatchView::OnMeasurementStart)
	ON_COMMAND(ID_MEASUREMENT_STOP, &CStopwatchView::OnMeasurementStop)
	ON_UPDATE_COMMAND_UI(ID_MEASUREMENT_START, &CStopwatchView::OnUpdateMeasurementStart)
	ON_UPDATE_COMMAND_UI(ID_MEASUREMENT_STOP, &CStopwatchView::OnUpdateMeasurementStop)
	ON_COMMAND(ID_MEASUREMENT_RESET, &CStopwatchView::OnMeasurementReset)
	ON_COMMAND(ID_MEASUREMENT_COPYHISTORYDATA, &CStopwatchView::OnMeasurementCopyhistorydata)
	ON_UPDATE_COMMAND_UI(ID_MEASUREMENT_COPYHISTORYDATA, &CStopwatchView::OnUpdateMeasurementCopyhistorydata)
END_MESSAGE_MAP()

// CStopwatchView construction/destruction

CStopwatchView::CStopwatchView()
{
	// TODO: add construction code here
	SampleRate=44100;	// default sample rate
}

CStopwatchView::~CStopwatchView()
{
}

BOOL CStopwatchView::PreCreateWindow(CREATESTRUCT& cs)
{
	// TODO: Modify the Window class or styles here by modifying
	//  the CREATESTRUCT cs

	return CView::PreCreateWindow(cs);
}

// CStopwatchView drawing

void CStopwatchView::OnDraw(CDC* xpDC)
{
	CStopwatchDoc* pDoc = GetDocument();
	ASSERT_VALID(pDoc);
	if (!pDoc)
		return;

	CRect w,w1,w2;
	GetClientRect(&w);

	CMemDC pDC(xpDC);

	CString s;
	int Length=10;
	CFont font;

	if (Counter/SampleRate>99) Length++;
	if (Counter/SampleRate>999) Length++;
	LOGFONT lf;
	memset(&lf, 0, sizeof(LOGFONT));       // zero out structure
	lf.lfHeight = 2*(w.right-w.left)/Length;
	if ((w.bottom-w.top)<lf.lfHeight)
	{
		lf.lfHeight=(w.bottom-w.top)-4;
	}
	strcpy(lf.lfFaceName, "Arial");        // request a face name "Arial"
	VERIFY(font.CreateFontIndirect(&lf));  // create the font
	CFont* def_font = pDC->SelectObject(&font);
	pDC->SetTextColor(RGB(0,0,0));
	w1=w2=w;
	w1.bottom=w2.top=(w.top+w.bottom)/2;
	s.Format(_T("T=%4.4lf s"),double(Counter)/SampleRate);
	pDC->DrawText(s,&w1,DT_LEFT | DT_SINGLELINE | DT_CENTER | DT_VCENTER);
	s.Format(_T("N=%d"),NumberOfEvents); //double(BlockingCounter)/SampleRate);
	pDC->DrawText(s,&w2,DT_LEFT | DT_SINGLELINE | DT_CENTER | DT_VCENTER);
	pDC->DrawEdge(&w,EDGE_SUNKEN,BF_RECT);
	pDC->SelectObject(def_font);
	font.DeleteObject();
}


// CStopwatchView printing

BOOL CStopwatchView::OnPreparePrinting(CPrintInfo* pInfo)
{
	// default preparation
	return DoPreparePrinting(pInfo);
}

void CStopwatchView::OnBeginPrinting(CDC* /*pDC*/, CPrintInfo* /*pInfo*/)
{
	// TODO: add extra initialization before printing
}

void CStopwatchView::OnEndPrinting(CDC* /*pDC*/, CPrintInfo* /*pInfo*/)
{
	// TODO: add cleanup after printing
}


// CStopwatchView diagnostics

#ifdef _DEBUG
void CStopwatchView::AssertValid() const
{
	CView::AssertValid();
}

void CStopwatchView::Dump(CDumpContext& dc) const
{
	CView::Dump(dc);
}

CStopwatchDoc* CStopwatchView::GetDocument() const // non-debug version is inline
{
	ASSERT(m_pDocument->IsKindOf(RUNTIME_CLASS(CStopwatchDoc)));
	return (CStopwatchDoc*)m_pDocument;
}
#endif //_DEBUG

// this function is called, when a buffer is filled (or some wave input operation occured)
void CALLBACK waveInHandler(HWAVEIN hwi, UINT uMsg, DWORD dwInstance, DWORD dwParam1, DWORD dwParam2)
{
	WAVEHDR *whdr = (WAVEHDR *)dwParam1;
	CStopwatchView *p = (CStopwatchView *)dwInstance;

	if (uMsg == WIM_DATA)	// data has been arrived
	{
		if (whdr->dwBytesRecorded==whdr->dwBufferLength)
		{
			if (whdr->lpData==WaveHeaderIn1.lpData)
			{
				waveInAddBuffer(hwi,&WaveHeaderIn1,sizeof(WaveHeaderIn1));	// add the next buffer to the queue
			}
			else 
			{
				waveInAddBuffer(HandleWaveIn,&WaveHeaderIn2,sizeof(WaveHeaderIn2));	// add the next buffer to the queue
			}
			short (*buffer)[2];
			buffer=(short (*)[2])whdr->lpData;

			for(int i=0; i<BufferLength; i++)
			{
				int x=-buffer[i][0];
				if (PreviousSample<=Threshold)
				{
					if (GateIsBlocked && x>Threshold+ThresholdTolerance)	// upward crossing
					{
						GateIsBlocked=0;
						if (TimerIsRunning) BlockingCounter=Counter;
					}
				}
				if (PreviousSample>=Threshold-ThresholdTolerance)
				{
					if (x<Threshold-ThresholdTolerance)
					{
						GateIsBlocked=1;
						TimerIsRunning=!TimerIsRunning;
						if (TimerIsRunning) Counter=0;
						else
						{
							if (NumberOfEvents<HISTORY_LENGTH) HistoryData[NumberOfEvents]=Counter;
							NumberOfEvents++;
						}
					}
				}
				if (TimerIsRunning)
				{
					Counter++;
				}
				PreviousSample=x;
			}
			p->Invalidate(0);
		}
	}
}

// CStopwatchView message handlers

void CStopwatchView::OnMeasurementStart()
{
	// TODO: Add your command handler code here
	BufferLength=0.183*SampleRate;	// update in every 183ms

	InputBuffer1 = new short[BufferLength][2];
	InputBuffer2 = new short[BufferLength][2];

	WaveInFormat.cbSize=0;
	WaveInFormat.nChannels=2;
	WaveInFormat.wBitsPerSample=16;
	WaveInFormat.nBlockAlign=WaveInFormat.nChannels*WaveInFormat.wBitsPerSample/8;
	WaveInFormat.nSamplesPerSec=SampleRate;
	WaveInFormat.nAvgBytesPerSec=WaveInFormat.nSamplesPerSec*WaveInFormat.nBlockAlign;
	WaveInFormat.wFormatTag=WAVE_FORMAT_PCM;
	waveInOpen(&HandleWaveIn, WAVE_MAPPER, &WaveInFormat, (DWORD_PTR)waveInHandler, (DWORD_PTR)this, CALLBACK_FUNCTION); //CALLBACK_NULL 
	WaveHeaderIn1.lpData=(LPSTR)InputBuffer1;
	WaveHeaderIn1.dwBufferLength=BufferLength*2*sizeof(short);
	WaveHeaderIn1.dwUser=0;
	WaveHeaderIn1.dwFlags=0;
	WaveHeaderIn2.lpData=(LPSTR)InputBuffer2;
	WaveHeaderIn2.dwBufferLength=BufferLength*2*sizeof(short);
	WaveHeaderIn2.dwUser=0;
	WaveHeaderIn2.dwFlags=0;

	waveInPrepareHeader(HandleWaveIn,&WaveHeaderIn1,sizeof(WaveHeaderIn1));
	waveInAddBuffer(HandleWaveIn,&WaveHeaderIn1,sizeof(WaveHeaderIn1));
	waveInPrepareHeader(HandleWaveIn,&WaveHeaderIn2,sizeof(WaveHeaderIn2));
	waveInAddBuffer(HandleWaveIn,&WaveHeaderIn2,sizeof(WaveHeaderIn2));

	Counter=BlockingCounter=0;
	PreviousSample=0;
	TimerIsRunning=0;
	GateIsBlocked=0;
	NumberOfEvents=0;

	MeasurementStarted=1;
	waveInStart(HandleWaveIn);
}

void CStopwatchView::OnMeasurementStop()
{
	// TODO: Add your command handler code here
	waveInStop(HandleWaveIn);
	waveInReset(HandleWaveIn);
	waveInUnprepareHeader(HandleWaveIn,&WaveHeaderIn1,sizeof(WaveHeaderIn1));
	waveInUnprepareHeader(HandleWaveIn,&WaveHeaderIn2,sizeof(WaveHeaderIn2));
	waveInClose(HandleWaveIn);

	delete [] InputBuffer1;
	delete [] InputBuffer2;
	MeasurementStarted=0;
}

void CStopwatchView::OnUpdateMeasurementStart(CCmdUI *pCmdUI)
{
	// TODO: Add your command update UI handler code here
	pCmdUI->Enable(!MeasurementStarted);
}

void CStopwatchView::OnUpdateMeasurementStop(CCmdUI *pCmdUI)
{
	// TODO: Add your command update UI handler code here
	pCmdUI->Enable(MeasurementStarted);
}

void CStopwatchView::OnMeasurementReset()
{
	// TODO: Add your command handler code here
	GateIsBlocked=0;
	Counter=BlockingCounter=0;
	NumberOfEvents=0;
	Invalidate();
}

void CStopwatchView::OnMeasurementCopyhistorydata()
{
	HANDLE h;                           
	LPSTR databuffer;
	char *dataptr;

	CString s; //("Event\tTime[s]\r\n");

	for(int i=0; i<NumberOfEvents; i++)
	{
		s.AppendFormat("%d\t%.4lf\r\n",i+1,double(HistoryData[i])/SampleRate);
	}

	char decimalpoint;
	GetLocaleInfo(LOCALE_USER_DEFAULT,LOCALE_SDECIMAL,&decimalpoint,1);	// what is the decimalpoint?
	if (decimalpoint!='.') s.Replace('.',decimalpoint);	// replace decimal point if needed

	if ( h=GlobalAlloc(GMEM_MOVEABLE,s.GetLength()))	// allocate memory for data
	{
		if (databuffer=(LPSTR)GlobalLock(h))			// lock this memory
		{
			memcpy(databuffer,s,s.GetLength());			// copy data to the memory 
			GlobalUnlock(h);							// unlock memory
			if (OpenClipboard())						// open the clipboard
			{
				EmptyClipboard();						// clear clipboard
				SetClipboardData(CF_TEXT,h);			// set clipboard contents
				CloseClipboard();						// close the clipboard
			}
			h=NULL;	
		}
		else
		{
			//AfxMessageBox("Can't lock memory",MB_OK | MB_ICONEXCLAMATION);
			if (GlobalFree(h)!=NULL)
			{
				//AfxMessageBox("Can't restore memory",MB_OK | MB_ICONEXCLAMATION);
			}	
		}               
	}
	else
	{
		//AfxMessageBox("Can't allocate memory",MB_OK | MB_ICONEXCLAMATION);
	}
}

void CStopwatchView::OnUpdateMeasurementCopyhistorydata(CCmdUI *pCmdUI)
{
	// TODO: Add your command update UI handler code here
	pCmdUI->Enable(NumberOfEvents);
}
void CStopwatchView::OnInitialUpdate()
{
	CStopwatchDoc* pDoc = GetDocument();
	pDoc->SetTitle("Time measurement");
	CView::OnInitialUpdate();

	// TODO: Add your specialized code here and/or call the base class
}
